use std::collections::HashMap;
use std::iter::Iterator;
use std::path::Path;
use vector::*;
use std::io;
use std;
use lwofront;
use ultraback;
use math;
use image;

type Image = image::Image;
type Vector3f = Vector3<f32>;

#[derive(Clone)]
struct Tile {
	p0: Vector2<isize>,
	p1: Vector2<isize>
}

struct MergeResult {
	vmap: lwofront::Vmap,
	pnts: lwofront::Pnts,
	pols: lwofront::Pols
}

struct Pts<'a> {
	point: &'a lwofront::Point,
	normal: Vector3f,
	tc_img: Option<(&'a [f32], &'a Image)>
}

fn split_layer_by_texture(layer: &lwofront::Layr, layers_base: u16, width: usize, height: usize) -> Vec<(lwofront::Layr, ultraback::Tile)> {
	let mut split = Vec::new();

	let vmap = match layer.find_vmap( lwofront::VmapType::TXUV ) {
		Some(vmap) => vmap,
		None => return split
	};

	for polygon in layer.get_polygons().get_vec() {			// czy tutaj polygon jest clonowany? jak otrzymac referencje?
		// zmapuj indeksy do PNTS starej warstwy na PNTS nowej
		let old_ids = polygon.get_verts();
		let mut idx_map = HashMap::new();
		for (new_idx, old_idx) in old_ids.iter().enumerate() {
			idx_map.insert( old_idx, new_idx );
		}

		let mut new_layer = lwofront::Layr::new(
			layers_base,
			0,
			&Vector3f::new( 0.0, 0.0, 0.0 ),
			"layer name",
			0
		);

		{
			let point_refs: Vec<&lwofront::Point> = layer.get_points().collect_refs(
					polygon.get_verts()
				);
			let mut points: &mut Vec<lwofront::Point> = new_layer.get_points_mut().get_vec_mut();
			for point in point_refs {
				points.push( point.clone() );
			}
		}

		let new_tile = {
			new_layer.get_vmaps_mut()
				.push(
					lwofront::Vmap::new(
						lwofront::VmapType::TXUV,
						2,
						"vmap name"
					)
				);

			let texcoord_refs: Vec<&lwofront::VmapViVal> = vmap.collect_refs(
					polygon.get_verts()
				);

			let mut texcoords: &mut Vec<lwofront::VmapViVal> = new_layer.get_vmaps_mut()[0].get_vec_mut();
			for tc in texcoord_refs {
				let tc_new = lwofront::VmapViVal {
					vi: *idx_map.get( &tc.vi ).unwrap() as u32,
					val: tc.val.clone()
				};

				texcoords.push( tc_new );
			}
			// min values
			let p0x = texcoords.iter()
				.map( |tc| -> f32 { tc.val[0] } )
				.fold( std::f32::NAN, f32::min ) as f32;
			let p0y = texcoords.iter()
				.map( |tc| -> f32 { tc.val[1] } )
				.fold( std::f32::NAN, f32::min ) as f32;
			// max values
			let p1x = texcoords.iter()
				.map( |tc| -> f32 { tc.val[0] } )
				.fold( std::f32::NAN, f32::max ) as f32;
			let p1y = texcoords.iter()
				.map( |tc| -> f32 { tc.val[1] } )
				.fold( std::f32::NAN, f32::max ) as f32;

			// this won't work properly with wrapping...
			ultraback::Tile::new(
				&Vector2::new(
					(width as f32 * p0x) as u16,
					(height as f32 * (1.0 - p1y)) as u16		// 1.0 - py for coordsys transformation
				),
				&Vector2::new(
					(width as f32 * p1x) as u16,
					(height as f32 * (1.0 - p0y)) as u16		// p1.y swapped with p0.y cause there must be: p0.y < p1.y
				)
			)
		};
		{
			let ids_new = polygon.get_verts().iter()
				.map( |old_idx| { *idx_map.get( old_idx ).unwrap() as u32 } )
				.collect();

			new_layer.get_polygons_mut()
				.get_vec_mut()
				.push(
					lwofront::Polygon::new (
						polygon.get_flags(),
						ids_new
					)
				);
		}
		
		split.push( (new_layer, new_tile) );
	}

	split
}

fn rgba8888_to_rgba5551(color: &image::PixelRGBA<u8>) -> u16 {
	((color.r as u16 & 0xF8) << 8)
	| ((color.g as u16 & 0xF8) << 3)
	| ((color.b as u16 & 0xF8) >> 2)
	| (color.a as u16 >> 7)
}

fn color_to_value( color: &image::PixelRGBA<u8> ) -> u32 {
	((color.r as u32) << 24) | ((color.g as u32) << 16)
	| ((color.b as u32) << 8) | (color.a as u32)
}

fn calculate_normal(vi: usize, points: &[lwofront::Point], polygons: &[lwofront::Polygon]) -> Vector3f {
	let acc_normal = polygons.iter()
		.filter( |&&ref face| {
			face.refers_to_vertex( vi as lwofront::Vx )
		} )
		.fold(
			(Vector3f::new( 0.0, 0.0, 0.0 ), 0),
			|acc, polygon| {
				let vts_ids = polygon.get_verts();
				let (p0,p1,p2) = (
					&points[vts_ids[0] as usize],
					&points[vts_ids[1] as usize],
					&points[vts_ids[2] as usize]
				);

				let u = math::sub_vec( p1, p0 );
				let v = math::sub_vec( p2, p0 );
				(
					math::add_vec(
						&acc.0,
						&math::normalize_vector(
							&math::cross_product( &u, &v )
						)
					),
					acc.1 + 1
				)
			}
		);

	math::scale_uniformly_vec( &acc_normal.0, 1.0 / acc_normal.1 as f32 )
}

// splits vertices affected by VMAD; updates polygons referring to these
fn merge_vmap_vmad(vmap: &lwofront::Vmap, vmad: &lwofront::Vmad, pols: &lwofront::Pols, pnts: &lwofront::Pnts) -> MergeResult {
	let mut result = MergeResult {
		vmap: vmap.clone(),
		pnts: pnts.clone(),
		pols: pols.clone()
	};

	for (vi, point) in pnts.get_vec().iter().enumerate() {
		// filter vmads referring to current point
		let vmad_v = vmad.get_vec().iter()
			.filter( |vi_pi_val| {
				vi_pi_val.vi == vi as u32
			} );

		for vi_pi_val in vmad_v {
			let new_vi = result.pnts.get_vec().len() as u32;
			result.pnts.add_point( &point );
			result.vmap.add_mapping(
				lwofront::VmapViVal {
					vi: new_vi,
					val: vi_pi_val.val.clone()
				}
			);

			let pi = vi_pi_val.pi as usize;
			result.pols.get_vec_mut()[pi].replace_vi( vi as u32, new_vi );
		}
	}

	result
}

fn convert_polygon(polygon: &lwofront::Polygon) -> ultraback::Face {
	let vts_ids = polygon.get_verts();
	match polygon.get_num_verts() {
		3 => ultraback::Face::Tri {
			ids: Vector3::<i16>::new(
				vts_ids[0] as i16,		// TODO: check that we can actualy cast without loss
				vts_ids[2] as i16,
				vts_ids[1] as i16
			)
		},
		4 => ultraback::Face::Quad {
			ids: Vector4::<i16>::new(
				vts_ids[1] as i16,
				vts_ids[0] as i16,
				vts_ids[3] as i16,
				vts_ids[2] as i16
			)
		},
		_ => panic!("Unsupported polygon"),			// TODO: don't panic, tessalate!
	}
}

fn convert_point(pnt: &Pts) -> ultraback::Vertex {
	let point = pnt.point;
	for coord in point {
		if ( coord < i16::min_value() as f32 ) || ( coord > i16::max_value() as f32 ) {
			println!("Warning: vertex coordinate doesn't fit in 16 bits")
		}
	}

	let texcoord_scale = 2*32;	// 32 for shifting 5 bits left, 2 for additional shift; TODO: this should be set by the user
	let tc = match pnt.tc_img {
		Some(tc_img) => {
			let mut tc = [0.0, 0.0];
			tc.clone_from_slice(tc_img.0);
			tc[0] *= tc_img.1.width as f32;
			tc[1] = 1.0 - tc[1];			// transform texture coordsys
			tc[1] *= tc_img.1.height as f32;

			if (tc[0].abs() as i16 > std::i16::MAX/(texcoord_scale as i16))
			|| (tc[1].abs() as i16 > std::i16::MAX/(texcoord_scale as i16)) {
				println!("Warning: texture coordinate out of range. Setting to 0.");
				tc[0] = 0.0;
				tc[1] = 0.0;
			}

			[tc[0], tc[1]]
		}
		None => [0.0, 0.0]
	};

	let normal = math::scale_uniformly_vec(&pnt.normal, 127.0);
	ultraback::Vertex::new(
		ultraback::Position::new(
			point.x() as i16,
			point.y() as i16,
			-point.z() as i16	// transform to RHS
		),
		ultraback::Texcoords::new(
			texcoord_scale * (tc[0] as i16),
		   	texcoord_scale * (tc[1] as i16)
		),
		ultraback::Normal::new(
			normal.x() as u8,
			normal.y() as u8,
			normal.z() as u8,
			0xFF
		)
	)
}

// NOTE: currently we accept only one surface per layer
fn convert_layer(
	src_layer: &lwofront::Layr,
	surf_id: usize,
	surfaces: &[&lwofront::Surf],
	textures: &[Option<Image>],
	attributes: &ultraback::LayerAttributes
) -> ultraback::Layer {

	let surface = surfaces[0];
	let vmad = src_layer.find_vmad( lwofront::VmadType::TXUV );

	// find TXUV-type VMAP and merge with VMAD (if both are present)
	let (vmap, pnts, pols) =
		match src_layer.find_vmap( lwofront::VmapType::TXUV ) {
			None => (None, src_layer.get_points().clone(), src_layer.get_polygons().clone() ),
			Some(vmap) => {
				let merged = match vmad {
					Some(vmad) => {
						merge_vmap_vmad(
							&vmap,
							&vmad,
							src_layer.get_polygons(),
							src_layer.get_points()
						)
					}
					None => {
						MergeResult {
							vmap: vmap.clone(),
							pnts: src_layer.get_points().clone(),
							pols: src_layer.get_polygons().clone()
						}
					}
				};

				(Some(merged.vmap), merged.pnts, merged.pols)
			}
		};

	let mut pts = Vec::<Pts>::new();
	for (vi, vertex) in pnts.get_vec().iter().enumerate() {
		// find texcoords
		let tc_img = match vmap {
			Some(ref vm) => {
				let tc = vm.get_vec().iter()
					.find( |&vi_val| {
						vi_val.vi == vi as u32
					} )
					.and_then( |ref vi_val| {
						Some(&vi_val.val)
					} ).unwrap();

				match surface.get_imap_texid() {
					Some(ti) => {
						match textures[ti as usize].as_ref() {
							Some(tex) => Some((tc.as_slice(), tex)),
							None => None
						}
					}
					_ => None
				}

			}
			None => None
		};

		pts.push( Pts {
			point: vertex,
			normal: calculate_normal( vi, pnts.get_vec(), pols.get_vec() ),
			tc_img: tc_img
		} );
	}

	// TODO: split vertex if polygons have different materials

	let mut verts = Vec::new();
	for pnt in &pts {
		verts.push(
			convert_point( &pnt )
		);
	}

	let mut faces = Vec::new();
	for polygon in pols.get_vec() {
		faces.push(
			convert_polygon( &polygon )
		);
	}

	let mut attr_copy = ultraback::LayerAttributes::new();
	attr_copy.clone_from( attributes );
	ultraback::Layer::new(
		verts,
		faces,
		surf_id,
		attr_copy
	)
}

fn convert_surface(src_surface: &lwofront::Surf) -> ultraback::Material {
	// assuming IMAP context
	let handle_blok_attributes = |attributes, mtl: &mut ultraback::Material| {
		for attribute in attributes {
			match attribute {
				&lwofront::BlokSubchunk::Imag { texture } => {
					mtl.set_tex_id( Some((texture-1)  as usize) );	// -1 !!!!!!
				}
				/*
				&lwofront::BlokSubchunk::Tmap { ref attributes } => {
					for attribute in attributes {
						match attribute {
							&lwofront::TmapSubchunk::Size { vector: ref v, .. } => {
								// TODO
							},
							_ => ()
						}
					}
				},
				&lwofront::BlokSubchunk::Axis { ref texture_axis } => {
					// TODO
				},
				*/
				_ => ()
			}
		}
	};

	let mut dst_mtl = ultraback::Material::new(
		None,
		Vector3f::new(0.5, 0.5, 0.5),
		1.0
	);
	for attribute in src_surface.get_attributes() {
		match attribute {
			&lwofront::SurfSubchunk::Colr { ref base, .. } => {
				dst_mtl.set_color( base );
			}
			&lwofront::SurfSubchunk::Blok { ref header, ref attributes } => {
				if let &lwofront::BlokHeader::Imap {..} = header {
					handle_blok_attributes( attributes, &mut dst_mtl );
				}
			}
			_ => ()
		}
	}

	dst_mtl
}

fn open_image(img_path: &str) -> Result<Image, io::Error> {
	image::Image::new( img_path )
}

fn convert_image(bitmap: &Image) -> ultraback::Texture {
	let mut histogram = HashMap::new();				// poor performance!
	let buffer: &Vec<_> = bitmap.pixels.as_ref();
	for i in 0..buffer.len() {
		*histogram.entry( buffer[i] ).or_insert( 0 ) += 1;
	}

	let mut texture = ultraback::Texture::new(
		bitmap.width as u16,
		bitmap.height as u16,

		if histogram.keys().len() <= 16 {
			ultraback::TextureData::CI4 {
				pixels: Vec::new(),
				palette: {
					let mut pal = ultraback::Palette::new();
					for color in histogram.keys() {
						pal.add_color(
							rgba8888_to_rgba5551( color )
						);
					}
					pal
				}
			}
		} else {
			ultraback::TextureData::RGBA16 { pixels: Vec::new() }
		}
	);

	// convert pixels
	match texture.get_data_mut() {
		&mut ultraback::TextureData::CI4 { ref mut pixels, ref palette } => {
			for i in 0.. {
				let colorl = match bitmap.pixels.get( 2*i ) {
					Some(c) => {
						palette.get_color_idx( rgba8888_to_rgba5551( c ) )
						.unwrap() as u8
					}
					None => break
				};
				let colorr = match bitmap.pixels.get( 2*i + 1 ) {
					Some(c) => {
						palette.get_color_idx( rgba8888_to_rgba5551( c ) )
						.unwrap() as u8
					}
					None => 0
				};
				pixels.push( (colorl << 4) | colorr );
			}
		}
		&mut ultraback::TextureData::RGBA16 { ref mut pixels } => {
			for i in 0.. {
				pixels.push( match bitmap.pixels.get( i ) {
					Some(c) => rgba8888_to_rgba5551( c ),
					None => break
				});
			}
		}
		&mut ultraback::TextureData::RGBA32 { ref mut pixels } => {
			for i in 0.. {
				pixels.push( match bitmap.pixels.get( i ) {
					Some(c) => color_to_value( c ),
					None => break
				});
			}
		}
		&mut ultraback::TextureData::Unknown { ref mut pixels } => {
			for i in 0.. {
				match bitmap.pixels.get( i ) {
					Some(c) => {
						pixels.push( c.r );
						pixels.push( c.g );
						pixels.push( c.b );
						pixels.push( c.a );
					}
					None => break
				}
			}
		}
	}

	texture
}

fn open_textures(clips: &[lwofront::Clip], src_mesh_dir: &str) -> Vec<Option<Image>> {
	let mut textures = Vec::new();
	for clip in clips {
		let mut imgpath = String::new();
		for subchunk in clip.get_subchunks() {
			if let &lwofront::ClipSubchunk::Stil { ref name } = subchunk {
				imgpath = src_mesh_dir.to_owned()
				+ "/"
				+ Path::new( name )
					.file_name()
					.unwrap()
					.to_str()
					.unwrap();
				break
			}
		};

		textures.push(
			match open_image( imgpath.as_str() ) {
				Ok(img) => Some(img),
				Err(_) => {
					println!("Warning: Couldn't open texture: {}", imgpath );
					None
				}
			}
		);
	}

	textures
}

pub fn convert(src_mesh: &lwofront::Object, src_mesh_dir: &str, generate_tiles: bool) -> Box<ultraback::Mesh> {
	let mut dst_mesh = Box::new( ultraback::Mesh::new() );

	// open textures and convert them
	let textures = open_textures( src_mesh.get_clips(), src_mesh_dir );
	for tex in &textures {
		let converted =
			if let &Some(ref tex) = tex {
				convert_image( &tex )
			} else {
				ultraback::Texture::new(
					0,
					0,
					ultraback::TextureData::Unknown { pixels: Vec::new() },
				)
			};

		if converted.get_texels_num() > converted.get_format().get_texels_limit() {
			println!( "Warning: texture won't fit into TMEM" ); //println!( "Warning: {}: {}", imgpath, warning ),
		}

		dst_mesh.add_texture( converted );
	}

	// convert surfaces
	for surface in src_mesh.get_surfaces() {
		dst_mesh.add_material(
			convert_surface( surface )
		);
	}

	// convert layers
	let mut layers_splitted_num = 0;
	let layers_base = src_mesh.get_layers().iter()
		.map( |layer| { layer.get_number() } ).max().unwrap();
	for (layer_idx, src_layer) in src_mesh.get_layers().iter().enumerate() {
		let surfaces_ids = src_mesh.get_surfaces_in_layer( src_layer );
		//let surfaces = src_mesh.collect_surfaces( &surfaces_ids );
		if surfaces_ids.len() > 1 {
			println!("Warning: more than one surface referenced in layer {}.", layer_idx);
		}
		let surf_id = surfaces_ids[0];

		let split: Option<Vec<_>> =
			if generate_tiles {
				if let Some(ti) = src_mesh.get_surfaces()[surf_id].get_imap_texid() {
					let tex = &dst_mesh.get_textures()[ti as usize];
					let lim = tex.get_format().get_texels_limit();

					if tex.get_texels_num() > lim {
						Some(split_layer_by_texture( src_layer, layers_base, tex.get_width(), tex.get_height() ))
					} else {
						None
					}
				} else {
					None
				}
			} else {
				None
			};
		
		let mut attributes = ultraback::LayerAttributes::new();
		if let Some(s) = split {
			attributes.insert( ultraback::LayerAttribute::PreventSeams );
			let mut i = 0;
			for &(ref l, ref t) in &s {
				dst_mesh.add_layer(
					convert_layer(
						&l,
						surf_id,
						&[&src_mesh.get_surfaces()[surf_id]],
						&textures,
						&attributes
					)
				);
				dst_mesh.add_layer_tile(
					layer_idx + layers_splitted_num + i,
					&t
				);
				i += 1;
			}
			layers_splitted_num += s.len();
		} else {
			dst_mesh.add_layer(
				convert_layer(
					src_layer,
					surf_id,
					&[&src_mesh.get_surfaces()[surf_id]],
					&textures,
					&attributes
				)
			);
		}
	}

	dst_mesh
}

