use std::fs::File;
use std::result;
use std::error;
use std::fmt;
use stream;
use vector;

type Vector3f = vector::Vector3<f32>;
type Result<T> = result::Result<T, Box<error::Error>>;
type Stream = stream::Stream<File>;
type Ordinal = Vec<u8>;
pub type Vx = u32;
pub type Point = Vector3f;

#[derive(Clone)]
pub struct Polygon {
	flags: u16,
	vtx_ids: Vec<Vx>
}

impl Polygon {
	pub fn new(flags: u16, vtx_ids: Vec<Vx>) -> Polygon {
		Polygon {
			flags: flags,
			vtx_ids: vtx_ids
		}
	}

	pub fn get_flags(&self) -> u16 {
		self.flags
	}

	pub fn get_verts(&self) -> &[Vx] {
		&self.vtx_ids
	}

	pub fn refers_to_vertex(&self, vx: Vx) -> bool {
		self.vtx_ids.iter().find(|vi: &&Vx| { **vi == vx }).is_some()
	}

	pub fn get_num_verts(&self) -> usize {
		self.vtx_ids.len()
	}

	pub fn replace_vi(&mut self, old_vi: Vx, new_vi: Vx) {
		let pvi = self.vtx_ids.iter().enumerate()
			.find( |&(_,vidx)| {
				*vidx == old_vi as u32
			} )
			.unwrap().0 as usize;
		self.vtx_ids[pvi] = new_vi;
	}
}

#[derive(Debug)]
enum LwoError {
	//Io(io::Error),
	Invalid,
	UnknownBlokHeader,
}

impl fmt::Display for LwoError {
	fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
		match *self {
			//LwoError::Io(ref err) => write!(f, "IO error: {}", err),
			LwoError::Invalid => write!( f, "Invalid" ),
			LwoError::UnknownBlokHeader => write!( f, "Unknown BLOK header" ),
		}
	}
}

impl error::Error for LwoError {
	fn description(&self) -> &str {
		match *self {
			//LwoError::Io(ref err) => err.description(),
			LwoError::Invalid => "IO error",
			LwoError::UnknownBlokHeader => "Unknown BLOK header",
		}
	}
	
	fn cause(&self) -> Option<&error::Error> {
		match *self {
			//LwoError::Io(ref err) => Some(err),
			LwoError::Invalid => None,
			LwoError::UnknownBlokHeader => None,
		}
	}
}

enum Literals {
	LWO2 = 0x4c574f32,
	TXUV = 0x54585556,
	SURF = 0x53555246,
	FACE = 0x46414354,
}

enum ChunkIDs {
	FORM = 0x464f524d,
	LAYR = 0x4c415952,
	PNTS = 0x504e5453,
	VMAP = 0x564d4150,
	VMAD = 0x564d4144,
	POLS = 0x504f4c53,
	TAGS = 0x54414753,
	PTAG = 0x50544147,
	CLIP = 0x434c4950,
	SURF = 0x53555246,
}
enum SubchunkIDs {
	STIL = 0x5354494c,
	COLR = 0x434f4c52,
	DIFF = 0x44494646,
	SPEC = 0x53504543,
	GLOS = 0x474c4f53,
	BLOK = 0x424c4f4b,
	IMAP = 0x494d4150,
	CHAN = 0x4348414e,
	ENAB = 0x454e4142,
	OPAC = 0x4f504143,
	//AXIS = 0x
	//NEGA = 0x4e454741,
	TMAP = 0x544d4150,
	CNTR = 0x434e5452,
	SIZE = 0x53495a45,
	ROTA = 0x524f5441,
	FALL = 0x46414c4c,
	OREF = 0x4f524546,
	CSYS = 0x43535953,
	PROJ = 0x50524f4a,
	AXIS = 0x41584953,
	IMAG = 0x494d4147,
	WRAP = 0x57524150,
	WRPW = 0x57525057,
	WRPH = 0x57525048,
	VMAP = 0x564d4150,
	AAST = 0x41415354,
	PIXB = 0x50495842,
}

struct Chunk {
	tag: u32,
	length: u32,
}

pub struct Object {
	tags: Tags,
	layers: Vec<Layr>,
	clips: Vec<Clip>,
	surfaces: Vec<Surf>,
}

pub struct Layr {
	number: u16,
	flags: u16,
	pivot: Vector3f,
	name: String,
	parent: u16,

	/* data from chunks belonging to layer */
	points: Pnts,
	vmaps: Vec<Vmap>,
	vmads: Vec<Vmad>,
	polygons: Pols,
	ptags: Ptag
}

#[derive(Clone)]
pub struct Pnts {
	v: Vec<Point>,
}

#[derive(Clone,PartialEq)]
pub enum VmapType {
	TXUV,
	Unknown(u32)
}
#[derive(Clone)]
pub struct VmapViVal {
	pub vi: Vx,
	pub val: Vec<f32>
}
#[derive(Clone)]
pub struct Vmap {
	typecode: VmapType,
	dimension: u16,
	name: String,
	v: Vec<VmapViVal>,
}

#[derive(Clone,PartialEq)]
pub enum VmadType {
	TXUV,
	Unknown(u32)
}

#[derive(Clone)]
pub struct VmadViPiVal {
	pub vi: Vx,
	pub pi: Vx,
	pub val: Vec<f32>,
}

#[derive(Clone)]
pub struct Vmad {
	pub typecode: VmadType,
	pub dimension: u16,
	pub name: String,
	v: Vec<VmadViPiVal>,
}

#[derive(Clone,Copy)]
pub enum PolsType {
	FACE,
	Unknown(u32)
}
#[derive(Clone)]
pub struct Pols {
	typecode: PolsType,
	v: Vec<Polygon>,
}
pub struct Tags {
	pub v: Vec<String>,
}

#[derive(Clone)]
pub enum PtagType {
	SURF,
	Unknown(u32)
}

#[derive(Clone)]
pub struct Ptag {
	pub typecode: PtagType,
	pub v: Vec<(Vx, u16)>,		// (POLS idx, TAGS idx)
}
pub enum ClipSubchunk {
	Stil { name: String },		// name doesn't contain '/0' at the end
	None,
}
pub struct Clip {
	index: u32,
	vsub: Vec<ClipSubchunk>,
}
pub enum SurfSubchunk {
	Colr { base: Vector3f, envelope: u32 },
	Diff { intensity: f32, envelope: u32 },
	Spec { intensity: f32, envelope: u32 },
	Glos { intensity: f32, envelope: u32 },
	Blok { header: BlokHeader, attributes: Vec<BlokSubchunk> },
}
pub struct Surf {
	name: String,
	source: String,
	attributes: Vec<SurfSubchunk>,
}

struct Subchunk {
	tag: u32,
	length: u16,
}
pub enum BlokHeader {
	Imap { ordinal: Ordinal, attributes: Vec<BlokHeaderSubchunk> },
	Proc { ordinal: Ordinal, attributes: Vec<BlokHeaderSubchunk> },
	Grad { ordinal: Ordinal, attributes: Vec<BlokHeaderSubchunk> },
	Shdr { ordinal: Ordinal, attributes: Vec<BlokHeaderSubchunk> },
}
pub enum BlokHeaderSubchunk {
	Chan { texture_channel: u32 },
	Enab { enable: u16 },
	Opac { typecode: u16, opacity: f32, envelope: Vx },
	Axis { displacement_axis: u16 }
}
pub enum BlokSubchunk {
	Tmap { attributes: Vec<TmapSubchunk> },
	Proj { mode: u16 },
	Axis { texture_axis: u16 },
	Imag { texture: Vx },
	Wrap { width: u16, height: u16 },
	Wrpw { cycles: f32, envelope: Vx },
	Wrph { cycles: f32, envelope: Vx },
	Vmap { txuv_map_name: String },
	Aast { flags: u16, strength: f32 },
	Pixb { flags: u16 },
}
pub enum TmapSubchunk {
	Cntr { vector: Vector3f, envelope: Vx },
	Size { vector: Vector3f, envelope: Vx },
	Rota { vector: Vector3f, envelope: Vx },
	Oref { object_name: String },
	Fall { typecode: u16, vector: Vector3f, envelope: Vx },
	Csys { typecode: u16 },
}

impl Object {
	pub fn new() -> Object {
		Object {
			tags: Tags { v: Vec::new() },
			layers: Vec::new(),
			clips: Vec::new(),
			surfaces: Vec::new(),
		}
	}
}

impl Layr {
	pub fn new(number: u16, flags: u16, pivot: &Vector3f, name: &str, parent: u16) -> Layr {
		Layr {
			number: number,
			flags: flags,
			pivot: pivot.clone(),
			name: String::from( name ),
			parent: parent,
			points: Pnts { v: Vec::new() },
			vmaps: Vec::new(),
			vmads: Vec::new(),
			polygons: Pols { typecode: PolsType::Unknown( 0 ), v: Vec::new() },
			ptags: Ptag { typecode: PtagType::Unknown( 0 ), v: Vec::new() }
		}
	}

	pub fn get_number(&self) -> u16 {
		self.number
	}

	pub fn get_points(&self) -> &Pnts {
		&self.points
	}

	pub fn get_points_mut(&mut self) -> &mut Pnts {
		&mut self.points
	}

	pub fn get_vmaps(&self) -> &Vec<Vmap> {
		&self.vmaps
	}

	pub fn get_vmaps_mut(&mut self) -> &mut Vec<Vmap> {
		&mut self.vmaps
	}

	pub fn find_vmap(&self, typecode: VmapType) -> Option<&Vmap> {
		self.vmaps.iter()
			.find( |vmap| {
				typecode == vmap.typecode
			} )
	}

	pub fn get_vmads(&self) -> &Vec<Vmad> {
		&self.vmads
	}

	pub fn get_vmads_mut(&mut self) -> &mut Vec<Vmad> {
		&mut self.vmads
	}

	pub fn find_vmad(&self, typecode: VmadType) -> Option<&Vmad> {
		self.vmads.iter()
			.find( |vmad| {
				typecode == vmad.typecode
			} )
	}

	pub fn get_polygons(&self) -> &Pols {
		&self.polygons
	}

	pub fn get_polygons_mut(&mut self) -> &mut Pols {
		&mut self.polygons
	}

	pub fn get_flags(&self) -> u16 {
		self.flags
	}

	pub fn get_pivot(&self) -> Vector3f {
		self.pivot.clone()
	}

	pub fn get_name(&self) -> &str {
		&self.name
	}

	pub fn get_parent(&self) -> u16 {
		self.parent
	}
}

impl Surf {
	pub fn get_name(&self) -> &String {
		&self.name
	}

	pub fn get_source(&self) -> &String {
		&self.source
	}

	pub fn get_attributes(&self) -> &Vec<SurfSubchunk> {
		&self.attributes
	}

	pub fn get_imap_texid(&self) -> Option<Vx> {
		// find tex index
		let get_texid = |blok_subchunks: &Vec<BlokSubchunk>| -> Option<Vx> {
			blok_subchunks.iter().find( |&bs| {
				match bs {
					&BlokSubchunk::Imag { .. } => true,
					_ => false
				}
			} ).and_then( |bs| {
				match bs {
					&BlokSubchunk::Imag { ref texture } => Some(*texture - 1),	// -1 !!!!!
					_ => None
				}
			} )
		};

		self.get_attributes().iter().find( |&attr| {
			match attr {
				&SurfSubchunk::Blok { ref header, .. } => {
					match header {
						&BlokHeader::Imap { .. } => true,
						_ => false
					}
				},
				_ => false
			}
		} ).and_then( |attr| {
			match attr {
				&SurfSubchunk::Blok { ref header, ref attributes } => {
					match header {
						&BlokHeader::Imap { .. } => get_texid( attributes ), 	
						_ => None
					}
				},
				_ => None
			}
		} )
	}
}

impl Clip {
	pub fn get_index(&self) -> u32 {
		self.index
	}

	pub fn get_subchunks(&self) -> &Vec<ClipSubchunk> {
		&self.vsub
	}
}

impl Vmap {
	pub fn new(typecode: VmapType, dimension: u16, name: &str) -> Vmap {
		Vmap {
			typecode: typecode,
			dimension: dimension,
			name: String::from( name ),
			v: Vec::new()
		}
	}

	pub fn add_mapping(&mut self, mapping: VmapViVal) {
		self.v.push( mapping );
	}

	pub fn get_vec(&self) -> &Vec<VmapViVal> {
		&self.v
	}

	pub fn get_vec_mut(&mut self) -> &mut Vec<VmapViVal> {
		&mut self.v
	}

	pub fn collect_refs(&self, ids: &[Vx]) -> Vec<&VmapViVal> {
		let mut refs = Vec::new();
		for i in ids {
			refs.push(
				if self.v[*i as usize].vi == *i {	// optimization to prevent linear search
					&self.v[*i as usize]
				} else {
					match self.v.iter().find( |vi_val| { vi_val.vi == *i } ) {
						Some(v) => v,
						None => continue
					}
				}
			)
		}

		refs
	}
}

impl Vmad {
	pub fn get_vec(&self) -> &Vec<VmadViPiVal> {
		&self.v
	}
}

impl Pnts {
	pub fn get_vec(&self) -> &Vec<Point> {
		&self.v
	}
	
	pub fn get_vec_mut(&mut self) -> &mut Vec<Point> {
		&mut self.v
	}

	pub fn add_point(&mut self, point: &Point) {
		self.v.push( point.clone() );
	}

	pub fn collect_refs(&self, ids: &[Vx]) -> Vec<&Point> {
		let mut refs = Vec::new();
		for i in ids {
			refs.push( &self.v[*i as usize] )
		}

		refs
	}
}

impl Pols {
	pub fn get_type(&self) -> PolsType {
		self.typecode
	}

	pub fn get_vec(&self) -> &Vec<Polygon> {
		&self.v
	}

	pub fn get_vec_mut(&mut self) -> &mut Vec<Polygon> {
		&mut self.v
	}
}

impl Object {
	pub fn get_layers(&self) -> &Vec<Layr> {
		&self.layers
	}

	pub fn get_layers_mut(&mut self) -> &mut Vec<Layr> {
		&mut self.layers
	}

	pub fn get_clips(&self) -> &Vec<Clip> {
		&self.clips
	}

	pub fn get_surfaces(&self) -> &Vec<Surf> {
		&self.surfaces
	}

	pub fn get_surfaces_in_layer(&self, layer: &Layr) -> Vec<usize> {
		self.get_surfaces().iter().enumerate()
		.filter( |&(_, ref surf)| {
			layer.ptags.v.iter()
			.find( |&&(_,tagidx)| {
				surf.get_name() == &self.tags.v[tagidx as usize]
			} ).is_some()
		} )
		.map( |(id, _)| { id } )
		.collect()
	}

	pub fn collect_surfaces(&self, ids: &[usize]) -> Vec<&Surf> {
		let mut refs = Vec::new();
		for i in ids {
			refs.push( &self.surfaces[*i] );
		}
		refs
	}
}

trait ChunkReader {
	fn read_chunk(&mut self) -> Result<Chunk>;
	fn read_subchunk(&mut self) -> Result<Subchunk>;
	fn read_vector3f(&mut self) -> Result<Vector3f>;
	fn read_string(&mut self) -> Result<String>;
	fn read_ordinal(&mut self) -> Result<Ordinal>;
	fn read_vx(&mut self) -> Result<u32>;
}
impl ChunkReader for Stream {
	fn read_chunk(&mut self) -> Result<Chunk> {
		let chunk = Chunk {
			tag: self.read_u32()?,
			length: self.read_u32()?,
		};

		Ok( chunk )
	}

	fn read_subchunk(&mut self) -> Result<Subchunk> {
		let subchunk = Subchunk {
			tag: self.read_u32()?,
			length: self.read_u16()?,
		};

		Ok( subchunk )
	}

	fn read_vector3f(&mut self) -> Result<Vector3f> {
		let v = Vector3f {
			v: [
				self.read_f32()?,
				self.read_f32()?,
				self.read_f32()?
			]
		};

		Ok( v )
	}

	fn read_string(&mut self) -> Result<String> {
		let mut s = Vec::<u8>::new() ;
		let bytes = self.read_until( '\0' as u8, &mut s )?;
		s.pop();		// delete '\0'
		if 1 == ( bytes & 0x1 ) {
			self.consume( 1 );
		}
		Ok( String::from_utf8( s )? )
	}

	fn read_ordinal(&mut self) -> Result<Ordinal> {
		let mut s = Vec::<u8>::new() ;
		let bytes = self.read_until( '\0' as u8, &mut s )?;
		if 1 == ( bytes & 0x1 ) {
			self.consume( 1 );
		}
		Ok( s )
	}

	fn read_vx(&mut self) -> Result<u32> {
		let val = self.read_u16()?;

		Ok( val as u32 )
	}
}

pub fn process(reader: &mut Stream) -> Result<Object> {
	let form = reader.read_chunk()?;
	match form.tag {
		id if id == ChunkIDs::FORM as u32 => handle_form( reader ),
		_ => Err( Box::new( LwoError::Invalid ) ),
	}
}

fn handle_form(reader: &mut Stream) -> Result<Object> {
	if reader.read_u32()? != Literals::LWO2 as u32 {
		return Err( Box::new( LwoError::Invalid ) )
	}

	let mut object = Object::new();

	while reader.limit() != 0 {
		let child: Chunk = reader.read_chunk()?;
		let old_limit = reader.limit;
		reader.limit = child.length as usize;

		match child.tag {
			id if id == ChunkIDs::LAYR as u32 => {
				object.layers.push( handle_layr( reader )? );	// borrow vs as_ref ?
			},
			id if id == ChunkIDs::PNTS as u32 => {
				object.layers.last_mut().unwrap().points = handle_pnts( reader )?;
			},
			id if id == ChunkIDs::VMAP as u32 => {
				object.layers.last_mut().unwrap().vmaps.push( handle_vmap( reader )? );
			},
			id if id == ChunkIDs::VMAD as u32 => {
				object.layers.last_mut().unwrap().vmads.push( handle_vmad( reader )? );
			},
			id if id == ChunkIDs::POLS as u32 => {
				object.layers.last_mut().unwrap().polygons = handle_pols( reader )?;
			},
			id if id == ChunkIDs::TAGS as u32 => {
				object.tags = handle_tags( reader )?;
			},
			id if id == ChunkIDs::PTAG as u32 => {
				object.layers.last_mut().unwrap().ptags = handle_ptag( reader )?;
			},
			id if id == ChunkIDs::CLIP as u32 => {
				object.clips.push( handle_clip( reader )? );
			},
			id if id == ChunkIDs::SURF as u32 => {
				object.surfaces.push( handle_surf( reader )? );
			},
			
			_ => reader.consume( child.length as usize ),
		};
		reader.limit = old_limit - child.length as usize;
	}

	Ok( object )
}

fn handle_layr(reader: &mut Stream) -> Result<Layr> {
	 Ok(
		Layr::new(
			reader.read_u16()?,			// czy kolejnosc ewaluacji jest gwarantowana???
			reader.read_u16()?,
			&reader.read_vector3f()?,
			&reader.read_string()?,
			reader.read_u16()?,			// TODO: this is optional!!!
		)
	)
}

fn handle_pnts(reader: &mut Stream) -> Result<Pnts> {
	let mut pnts = Pnts { v: Vec::new() };
	while reader.limit() != 0  {
		pnts.v.push( reader.read_vector3f()? );
	}

	Ok( pnts )
}

fn handle_vmap(reader: &mut Stream) -> Result<Vmap> {
	let map_type = |t: u32| -> VmapType {
		match t {
			t if t == Literals::TXUV as u32 => VmapType::TXUV,
			_ => VmapType::Unknown( t )
		}
	};

	let mut vmap = Vmap::new(
		map_type( reader.read_u32()? ),
		reader.read_u16()?,
		&reader.read_string()?
	);

	while reader.limit() != 0 {
		let vx = reader.read_vx()?;
		let mut v = Vec::new();
		for _ in 0..vmap.dimension {
			v.push( reader.read_f32()? );
		}
		vmap.v.push( VmapViVal { vi: vx, val: v } );
	}

	Ok(vmap)
}

fn handle_vmad(reader: &mut Stream) -> Result<Vmad> {
	let map_type = |t: u32| -> VmadType {
		match t {
			t if t == Literals::TXUV as u32 => VmadType::TXUV,
			_ => VmadType::Unknown( t )
		}
	};

	let mut vmad = Vmad {
		typecode: map_type( reader.read_u32()? ),
		dimension: reader.read_u16()?,
		name: reader.read_string()?,
		v: Vec::new()
	};

	while reader.limit() != 0 {
		let vx = reader.read_vx()?;
		let px = reader.read_vx()?;
		let mut v = Vec::new();
		for _ in 0..vmad.dimension {
			v.push( reader.read_f32()? );
		}
		vmad.v.push( VmadViPiVal { vi: vx, pi: px, val: v } );
	}

	Ok( vmad )
}

fn handle_pols(reader: &mut Stream) -> Result<Pols> {
	let map_type = |t: u32| -> PolsType {
		match t {
			t if t == Literals::FACE as u32 => PolsType::FACE,
			_ => PolsType::Unknown( t )
		}
	};

	let mut pols = Pols {
		typecode: map_type( reader.read_u32()? ),
		v: Vec::new(),
	};

	let mut flags;
	let mut n;
	while reader.limit() != 0 {
		n = reader.read_u16()?;		// TODO: mask
		flags = n & (0x3F << 10);		// TODO: mask
		n &= 0x3FF;

		let mut vtx_ids = Vec::new();
		for _ in 0..n {
			vtx_ids.push( reader.read_vx()? );
		}

		pols.v.push(
			Polygon::new(flags, vtx_ids)
		);
	}

	Ok( pols )
}

fn handle_tags(reader: &mut Stream) -> Result<Tags> {
	let mut tags = Tags { v: Vec::new() };

	while reader.limit() != 0 {
		tags.v.push( reader.read_string()? );
	}

	Ok(tags)
}

fn handle_ptag(reader: &mut Stream) -> Result<Ptag> {
	let map_type = |t: u32| -> PtagType {
		match t {
			t if t == Literals::SURF as u32 => PtagType::SURF,
			_ => PtagType::Unknown( t )
		}
	};

	let mut ptag = Ptag {
		typecode: map_type( reader.read_u32()? ),
		v: Vec::new(),
	};

	while reader.limit() != 0 {
		ptag.v.push( (reader.read_u16()? as u32, reader.read_u16()?) );
	}

	Ok( ptag )
}

fn handle_clip(reader: &mut Stream) -> Result<Clip> {
	let mut clip = Clip {
		index: reader.read_u32()?,
		vsub: Vec::new(),
	};


	while reader.limit() != 0 {
		let subchunk = reader.read_subchunk()?;
		let old_limit = reader.limit;
		reader.limit = subchunk.length as usize;

		match subchunk.tag {
			id if id == SubchunkIDs::STIL as u32 => {
				clip.vsub.push( handle_sub_stil( reader)? );
			},
			_ => reader.consume( subchunk.length as usize ),
		}
		reader.limit = old_limit - subchunk.length as usize;
	}

	Ok( clip )
}

fn handle_sub_stil(reader: &mut Stream) -> Result<ClipSubchunk> {
	Ok( ClipSubchunk::Stil {
		name: reader.read_string()?
	} )
}

fn handle_surf(reader: &mut Stream) -> Result<Surf> {
	let mut surf = Surf {
		name: reader.read_string()?,
		source: reader.read_string()?,
		attributes: Vec::new()
	};


	while reader.limit() != 0 {
		let subchunk = reader.read_subchunk()?;
		let old_limit = reader.limit;
		reader.limit = subchunk.length as usize;

		match subchunk.tag {
			id if id == SubchunkIDs::COLR as u32 => {
				surf.attributes.push( handle_sub_colr(reader)? );
			},
			id if id == SubchunkIDs::DIFF as u32 => {
				surf.attributes.push( handle_sub_diff(reader)? );
			},
			id if id == SubchunkIDs::SPEC as u32 => {
				surf.attributes.push( handle_sub_spec(reader)? );
			},
			id if id == SubchunkIDs::GLOS as u32 => {
				surf.attributes.push( handle_sub_glos(reader)? );
			},
			id if id == SubchunkIDs::BLOK as u32 => {
				surf.attributes.push( handle_sub_blok(reader)? );
			},
			_ => reader.consume(subchunk.length as usize),
		}
		reader.limit = old_limit - subchunk.length as usize;
	}

	Ok( surf )
}

fn handle_sub_colr(reader: &mut Stream) -> Result<SurfSubchunk> {
	Ok( SurfSubchunk::Colr {
		base: reader.read_vector3f()?,
		envelope: reader.read_vx()?
	} )
}

fn handle_sub_diff(reader: &mut Stream) -> Result<SurfSubchunk> {
	Ok( SurfSubchunk::Diff {
		intensity: reader.read_f32()?,
		envelope: reader.read_vx()?
	} )
}

fn handle_sub_spec(reader: &mut Stream) -> Result<SurfSubchunk> {
	Ok( SurfSubchunk::Spec {
		intensity: reader.read_f32()?,
		envelope: reader.read_vx()?
	} )
}

fn handle_sub_glos(reader: &mut Stream) -> Result<SurfSubchunk> {
	Ok( SurfSubchunk::Glos {
		intensity: reader.read_f32()?,
		envelope: reader.read_vx()?
	} )
}

fn handle_sub_blok(reader: &mut Stream) -> Result<SurfSubchunk> {
	let subchunk = reader.read_subchunk()?;
	let old_limit = reader.limit;
	reader.limit = subchunk.length as usize;

	let mut blok = SurfSubchunk::Blok {
		header: match subchunk.tag {
			id if id == SubchunkIDs::IMAP as u32 => {
				handle_sub_imap( reader )?
			},
			_ => return Err( Box::new( LwoError::UnknownBlokHeader ) )
		},
		attributes: Vec::new()
	};
	reader.limit = old_limit - subchunk.length as usize;
	
	while reader.limit() != 0 {
		let subchunk = reader.read_subchunk()?;
		let old_limit = reader.limit;
		reader.limit = subchunk.length as usize;

		match subchunk.tag {
			id if id == SubchunkIDs::TMAP as u32 => {
				match &mut blok {
					&mut SurfSubchunk::Blok { header: _, attributes: ref mut attr } =>
						attr.push( handle_sub_tmap( reader )? ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::PROJ as u32 => {
				match &mut blok {
					&mut SurfSubchunk::Blok { header: _, attributes: ref mut attr } =>
						attr.push( handle_sub_proj( reader )? ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::AXIS as u32 => {
				match &mut blok {
					&mut SurfSubchunk::Blok { header: _, attributes: ref mut attr } =>
						attr.push( handle_sub_axis( reader )? ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::IMAG as u32 => {
				match &mut blok {
					&mut SurfSubchunk::Blok { header: _, attributes: ref mut attr } =>
						attr.push( handle_sub_imag( reader )? ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::WRAP as u32 => {
				match &mut blok {
					&mut SurfSubchunk::Blok { header: _, attributes: ref mut attr } =>
						attr.push( handle_sub_wrap( reader )? ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::WRPW as u32 => {
				match &mut blok {
					&mut SurfSubchunk::Blok { header: _, attributes: ref mut attr } =>
						attr.push( handle_sub_wrpw( reader )? ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::WRPH as u32 => {
				match &mut blok {
					&mut SurfSubchunk::Blok { header: _, attributes: ref mut attr } =>
						attr.push( handle_sub_wrph( reader )? ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::VMAP as u32 => {
				match &mut blok {
					&mut SurfSubchunk::Blok { header: _, attributes: ref mut attr } =>
						attr.push( handle_sub_vmap( reader )? ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::AAST as u32 => {
				match &mut blok {
					&mut SurfSubchunk::Blok { header: _, attributes: ref mut attr } =>
						attr.push( handle_sub_aast( reader )? ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::PIXB as u32 => {
				match &mut blok {
					&mut SurfSubchunk::Blok { header: _, attributes: ref mut attr } =>
						attr.push( handle_sub_pixb( reader )? ),
					_ => (),
				};
			},
			_ => reader.consume( subchunk.length as usize ),
		}
		reader.limit = old_limit - subchunk.length as usize;
	}

	Ok( blok )
}

fn handle_sub_imap(reader: &mut Stream) -> Result<BlokHeader> {
	let mut imap = BlokHeader::Imap {
		ordinal: reader.read_ordinal()?,
		attributes: Vec::new()
	};
	
	while reader.limit() != 0 {
		let subchunk = reader.read_subchunk()?;
		let old_limit = reader.limit;
		reader.limit = subchunk.length as usize;

		match subchunk.tag {
			id if id == SubchunkIDs::CHAN as u32 => {
				match &mut imap {
					&mut BlokHeader::Imap { ordinal: _, attributes: ref mut attr } =>
						attr.push( handle_sub_chan( reader )? ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::ENAB as u32 => {
				match &mut imap {
					&mut BlokHeader::Imap { ordinal: _, attributes: ref mut attr } =>
						attr.push( handle_sub_enab( reader )? ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::OPAC as u32 => {
				match &mut imap {
					&mut BlokHeader::Imap { ordinal: _, attributes: ref mut attr } =>
						attr.push( handle_sub_opac( reader )? ),
					_ => (),
				};
			},
			_ => reader.consume( subchunk.length as usize ),
		}
		reader.limit = old_limit - subchunk.length as usize;
	}

	Ok( imap )
}

fn handle_sub_chan(reader: &mut Stream) -> Result<BlokHeaderSubchunk> {
	let chan = BlokHeaderSubchunk::Chan {
		texture_channel: reader.read_u32()?,
	};

	Ok( chan )
}

fn handle_sub_enab(reader: &mut Stream) -> Result<BlokHeaderSubchunk> {
	let enab = BlokHeaderSubchunk::Enab {
		enable: reader.read_u16()?,
	};

	Ok( enab )
}

fn handle_sub_opac(reader: &mut Stream) -> Result<BlokHeaderSubchunk> {
	let opac = BlokHeaderSubchunk::Opac {
		typecode: reader.read_u16()?,
		opacity: reader.read_f32()?,
		envelope: reader.read_vx()?,
	};

	Ok( opac )
}
/*
fn handle_sub_axis(reader: &mut Stream) -> Result<BlokHeaderSubchunk> {
	let axis = BlokHeaderSubchunk::Axis {
		displacement_axis: reader.read_u16()?,
	};

	Ok(axis)
}
*/
fn handle_sub_tmap(reader: &mut Stream) -> Result<BlokSubchunk> {
	let mut tmap = BlokSubchunk::Tmap {
		attributes: Vec::new()
	};

	while reader.limit() != 0 {
		let subchunk = reader.read_subchunk()?;
		let old_limit = reader.limit;
		reader.limit = subchunk.length as usize;

		match subchunk.tag {
			id if id == SubchunkIDs::CNTR as u32 => {
				match &mut tmap {
					&mut BlokSubchunk::Tmap { attributes: ref mut attr } =>
						attr.push( handle_sub_cntr( reader )? ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::SIZE as u32 => {
				match &mut tmap {
					&mut BlokSubchunk::Tmap { attributes: ref mut attr } =>
						attr.push( handle_sub_size( reader )? ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::ROTA as u32 => {
				match &mut tmap {
					&mut BlokSubchunk::Tmap { attributes: ref mut attr } =>
						attr.push( handle_sub_rota( reader )? ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::OREF as u32 => {
				match &mut tmap {
					&mut BlokSubchunk::Tmap { attributes: ref mut attr } =>
						attr.push( handle_sub_oref( reader )? ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::FALL as u32 => {
				match &mut tmap {
					&mut BlokSubchunk::Tmap { attributes: ref mut attr } =>
						attr.push( handle_sub_fall(reader )? ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::CSYS as u32 => {
				match &mut tmap {
					&mut BlokSubchunk::Tmap { attributes: ref mut attr } =>
						attr.push( handle_sub_csys( reader )? ),
					_ => (),
				};
			},
			_ => reader.consume( subchunk.length as usize ),
		}
		reader.limit = old_limit - subchunk.length as usize;
	}

	Ok( tmap )
}

fn handle_sub_cntr(reader: &mut Stream) -> Result<TmapSubchunk> {
	Ok( TmapSubchunk::Cntr {
		vector: reader.read_vector3f()?,
		envelope: reader.read_vx()?,
	} )
}

fn handle_sub_size(reader: &mut Stream) -> Result<TmapSubchunk> {
	Ok( TmapSubchunk::Size {
		vector: reader.read_vector3f()?,
		envelope: reader.read_vx()?,
	} )
}

fn handle_sub_rota(reader: &mut Stream) -> Result<TmapSubchunk> {
	Ok( TmapSubchunk::Rota {
		vector: reader.read_vector3f()?,
		envelope: reader.read_vx()?,
	} )
}

fn handle_sub_oref(reader: &mut Stream) -> Result<TmapSubchunk> {
	Ok( TmapSubchunk::Oref {
		object_name: reader.read_string()?,
	} )
}

fn handle_sub_fall(reader: &mut Stream) -> Result<TmapSubchunk> {
	Ok( TmapSubchunk::Fall {
		typecode: reader.read_u16()?,
		vector: reader.read_vector3f()?,
		envelope: reader.read_vx()?,
	} )
}

fn handle_sub_csys(reader: &mut Stream) -> Result<TmapSubchunk> {
	Ok( TmapSubchunk::Csys {
		typecode: reader.read_u16()?,
	} )
}

fn handle_sub_proj(reader: &mut Stream) -> Result<BlokSubchunk> {
	Ok( BlokSubchunk::Proj {
		mode: reader.read_u16()?,
	} )
}

fn handle_sub_axis(reader: &mut Stream) -> Result<BlokSubchunk> {
	Ok( BlokSubchunk::Axis {
		texture_axis: reader.read_u16()?,
	} )
}

fn handle_sub_imag(reader: &mut Stream) -> Result<BlokSubchunk> {
	Ok( BlokSubchunk::Imag {
		texture: reader.read_vx()?,
	} )
}

fn handle_sub_wrap(reader: &mut Stream) -> Result<BlokSubchunk> {
	Ok( BlokSubchunk::Wrap {
		width: reader.read_u16()?,
		height: reader.read_u16()?,
	} )
}

fn handle_sub_wrpw(reader: &mut Stream) -> Result<BlokSubchunk> {
	Ok( BlokSubchunk::Wrpw {
		cycles: reader.read_f32()?,
		envelope: reader.read_vx()?,
	} )
}

fn handle_sub_wrph(reader: &mut Stream) -> Result<BlokSubchunk> {
	Ok( BlokSubchunk::Wrpw {
		cycles: reader.read_f32()?,
		envelope: reader.read_vx()?,
	} )
}

fn handle_sub_vmap(reader: &mut Stream) -> Result<BlokSubchunk> {
	Ok( BlokSubchunk::Vmap {
		txuv_map_name: reader.read_string()?,
	} )
}

fn handle_sub_aast(reader: &mut Stream) -> Result<BlokSubchunk> {
	Ok( BlokSubchunk::Aast {
		flags: reader.read_u16()?,
		strength: reader.read_f32()?,
	} )
}

fn handle_sub_pixb(reader: &mut Stream) -> Result<BlokSubchunk> {
	Ok( BlokSubchunk::Pixb {
		flags: reader.read_u16()?,
	} )
}
